#include "FlowGeometry.h"
#include "reverseBytes.h"

bool FlowGeometry::readFromFile(char* header, FILE* fp, bool bigEndian)
{
	isFlipped = false;
	//determine the dimensions
    sscanf(header,"SN4DB %d %d %d",&dim[0],&dim[1],&dim[2]);
    std::cout << "Dimensions: " << dim[0] << " x " << dim[1] << " x " << dim[2] << std::endl;
    if (dim[2]!=1)
    {
        std::cerr << "Invalid Z dimension value." << std::endl;
        return false;
    }

	geometryData = new vec3[dim[0]*dim[1]];
	//read the data and check if everything went fine
    int result = fread(geometryData,sizeof(vec3),dim[0]*dim[1],fp);
	if (result != dim[0]*dim[1])
	{
		std::cerr << "+ Error reading grid file." << std::endl << std::endl;
		return false;
	}

	if (bigEndian)
		for (int j = 0; j < getDimX()*getDimY(); j++)
			for (int k = 0; k < 3; k++)
				geometryData[j][k] = reverseBytes<float>(geometryData[j][k]);

    //first vertex
	boundaryMin = vec3(getPos(0));
	//last vertex
	boundaryMax = vec3(getPos((dim[0]*dim[1]) - 1));
	boundarySize = boundaryMax - boundaryMin;

    //if X and Y are swapped... meaning that going in a row the y value increases and x stays the same
    if (getPosY(dim[0]-1)>(boundaryMin[1] + boundaryMax[1])*0.5)
    {
		//this flag will fix that, anywhere where it's neccerasy... watch it :)
        isFlipped = true;
        std::cout << "Flipped Y and X dimensions." << std::endl;
    }
    else isFlipped = false;        

    std::cout << "X Boundaries: " << boundaryMin[0] << " ... " << boundaryMax[0] << std::endl;
    std::cout << "Y Boundaries: " << boundaryMin[1] << " ... " << boundaryMax[1] << std::endl;
    
    return true;
}

FlowGeometry::FlowGeometry()
{
    geometryData = NULL;
	isFlipped = false;
}

FlowGeometry::~FlowGeometry()
{
    if (geometryData)
        delete[] geometryData;
}

///returns X index of the last vertex lying left to the position x and the Y index of the last vertex lying under the position y 
int FlowGeometry::getXYvtx(vec3 pos)
{
    int i;
	int j;
	//search for the column left to the vertex
    for (i = 0; (i < dim[0])&&(getPosX(getVtx(i,0)) < pos[0]); i++);
	//search for the row under the vertex
	for (j = 0; (j < dim[1])&&(getPosY(getVtx(0,j)) < pos[1]); j++);
	
	//return the vertex ID of the found vertex
	return getVtx((i<dim[0]) ? i : dim[0]-1, (j<dim[1]) ? j : dim[1]-1);
}

inline vec3 FlowGeometry::getPos(int vtxID)
{
	return geometryData[vtxID];
}

inline float FlowGeometry::getPosX(int vtxID)
{
	return geometryData[vtxID][0];
}

inline float FlowGeometry::getPosY(int vtxID)
{
	return geometryData[vtxID][1];
}

//TODO students: improve this
///a very slow and dumb routine, that finds the nearest vertex to the given position
int FlowGeometry::getNearestVtx(vec3 pos)
{
	//take the norm to the first vertex
	float dist = pos.dist2(getPos(0));
	//mark the vertex 0 as the closest one
	int closest = 0;

	int next = getRightNeigh(0);
	float newd = pos.dist2(getPos(next));
	
	//Iterate through vertices from top left corner to bottom right and search for the nearest one.
	//Move to from top left corner to the right until distance starts increasing again.
	while (newd <= dist)
	{	
		closest = next;
		dist = newd;
		next = getRightNeigh(closest);
		newd = pos.dist2(getPos(next));
	}

	next = getBottomNeigh(closest);
	newd = pos.dist2(getPos(next));

	//Now move downward until distance starts increasing again. The closest point should
	//have been found now.
	while (newd <= dist)
	{	
		closest = next;
		dist = newd;
		next = getBottomNeigh(closest);
		newd = pos.dist2(getPos(next));
	}
	return closest;
}

bool FlowGeometry::getInterpolationAt(vec3 pos, int* vtxID, float* coef)
{
    //if we are outside of the dataset, return false
	//please note, that this test is valid only for rectangular datasets (block, hurricane), but not for tube
    if ((pos[0]<boundaryMin[0])||(pos[1]<boundaryMin[1])||(pos[0]>boundaryMax[0])||(pos[1]>boundaryMax[1]))
        return false;


	vtxID[0] = getNearestVtx(pos); //finds the nearest vertex to the given position and returns it's value as the dominanting one.

	//Now the other three closest neigbors are added. The closest 4 neighbors form a
	//rectangle around pos, so depending on which corner i'm at (4 possibilities),  i
	//save its remaining 3 neighbors.
	float dX = pos[0] - getPos(vtxID[0])[0];
    float dY = pos[1] - getPos(vtxID[0])[1];

    if (dX > 0) {							//Closest vertex is on the left
        vtxID[1] = getRightNeigh(vtxID[0]);
    } else {
        vtxID[1] = getLeftNeigh(vtxID[0]);	//Closest vertex is on the right or pos is right on
											//the line between two vertices
        if (vtxID[1] < 0) vtxID[1] = getRightNeigh(vtxID[0]); //closest has no left neighbor
    }

    if (dY > 0) {						 //Closest vertex is on the top
        vtxID[2] = getBottomNeigh(vtxID[0]);
        vtxID[3] = getBottomNeigh(vtxID[1]);
    } else {							
        vtxID[2] = getTopNeigh(vtxID[0]); //Closest vertex is on the bottom or pos is right on
        vtxID[3] = getTopNeigh(vtxID[1]); //the line between two vertices
        if (vtxID[2] < 0) {
            vtxID[2] = getBottomNeigh(vtxID[0]); //closest has no top neighbor
            vtxID[3] = getBottomNeigh(vtxID[1]);
        }
    }

	//Now we need to calculate the coefficients corresponding to the distance from each point
	//to pos. Closer points should have stronger influence.
	float gridDiffX = getPos(vtxID[1])[0]-getPos(vtxID[0])[0];
    float gridDiffY = getPos(vtxID[2])[1]-getPos(vtxID[0])[1];

    float ratio[4];
    ratio[0] = (gridDiffX-dX)/gridDiffX;
    ratio[1] = dX/gridDiffX;
    ratio[2] = (gridDiffY-dY)/gridDiffY;
    ratio[3] = dY/gridDiffY;
    coef[0] = ratio[0]*ratio[2];
    coef[1] = ratio[1]*ratio[2];
    coef[2] = ratio[0]*ratio[3];
    coef[3] = ratio[1]*ratio[3];

    return true;
}

inline float FlowGeometry::getMinX()
{
    return boundaryMin[0];
}

inline float FlowGeometry::getMaxX()
{
    return boundaryMax[0];
}

inline float FlowGeometry::getMinY()
{
    return boundaryMin[1];
}

inline float FlowGeometry::getMaxY()
{
    return boundaryMax[1];
}

inline int FlowGeometry::getDimX()
{
	return (isFlipped) ? dim[1] : dim[0];
}

inline int FlowGeometry::getDimY()
{
	return (isFlipped) ? dim[0] : dim[1];
}

inline int FlowGeometry::getDimZ()
{
    return dim[2];
}

int FlowGeometry::getVtx(int x, int y)
{
	//if we need to flip the rows and columns, we do it here 
	return (isFlipped)? (x*dim[1]) + y : (y*dim[0]) + x;
}

int FlowGeometry::getVtxX(int vtxID)
{
	//if we need to flip the rows and columns, we do it here
	return (isFlipped)? vtxID / dim[1] : vtxID % dim[0];
}

int FlowGeometry::getVtxY(int vtxID)
{
	//if we need to flip the rows and columns, we do it here
	return (isFlipped)? vtxID % dim[1] : vtxID / dim[0];
}

int FlowGeometry::getRightNeigh(int vtxID)
{
    int x = getVtxX(vtxID);
    return (x+1 < dim[0]) ? getVtx(x+1,getVtxY(vtxID)) : -1;
}

int FlowGeometry::getTopNeigh(int vtxID)
{
	//remember that the data is structured with (0,0) in the upper-left corner and (1,1) in the lower-right
	//that's why we are subtracting 1 to find the top neighbour
    int y = getVtxY(vtxID);
    return (y > 1) ? getVtx(getVtxX(vtxID), y-1) : -1;
}

int FlowGeometry::getLeftNeigh(int vtxID)
{
    int x = getVtxX(vtxID);
    return (x > 1) ? getVtx(x-1,getVtxY(vtxID)) : -1;
}

int FlowGeometry::getBottomNeigh(int vtxID)
{
	//remember that the data is structured with (0,0) in the upper-left corner and (1,1) in the lower-right
    int y = getVtxY(vtxID);
    return (y+1 < dim[1]) ? getVtx(getVtxX(vtxID), y+1) : -1;
}

vec3 FlowGeometry::normalizeCoords(vec3 pos)
{
	vec3 u(pos - boundaryMin);
	//scale each component according to the side length
	u[0] /= boundarySize[0];
	u[1] /= boundarySize[1];

	return u;
}

vec3 FlowGeometry::unNormalizeCoords(vec3 pos)
{
	vec3 u(pos);
	//multiply each component according to the side length
	u[0] *= boundarySize[0];
	u[1] *= boundarySize[1];

	return u += boundaryMin;
}

bool FlowGeometry::getIsFlipped()
{
	return isFlipped;
}