#include "Streamlines.h"

/**
* Streamlines constructor.
* This class calculates Streamlines from a given 'dataset'. Therefore a grid is constructed with a given
* 'distance' between its points. From each gridpoint a line of length 'stepSize' is drawn in direction of
* the vector at that point, followd by another...until 'maxSteps' is reached.
*
* @param FlowData dataset
* @param int distance
* @param float stepSize
* @param int maxSteps
* @param float width
*
*/
Streamlines::Streamlines (FlowData* dataset, int distance, float stepSize, int maxSteps, float width)
{
	dist = distance;
	step = stepSize;
	max  = maxSteps;
	data = dataset;
	w = width;
}

/**
* Create grid for streamline construction.
* The grid points serve as seeds from which the streamlines are drawn following the 
* velocity vectors of the data set. The velocity vectors are followed backwards too,
* to make sure the lines are drawn all the way to the edges. The parameter 'formula'
* specifies if Euler or Runge Kutta is used to calculate the line.
*
* @param int formula
*
*/
void Streamlines::createGrid(int formula)
{
	float pointX;
	float pointY;
	float diffX = (float)data->getChannel(0)->geom->getDimX()/w;

	for(int i=dist; i < data->getChannel(0)->geom->getDimX(); i=i+dist)
	{
		for(int j=dist; j < data->getChannel(0)->geom->getDimY(); j=j+dist)
		{
			pointX = ((((float)i/data->getChannel(0)->geom->getDimX())*2)-1)*diffX;
			pointY = ((((float)j/data->getChannel(0)->geom->getDimY())*2)-1);

			//Calculate direction and follow step size .. till maxSteps
			if(formula == 1)
				getEulerPoints((pointX/diffX+1)*0.5, (pointY+1)*0.5);
			else if(formula == 2)
				getRungeKuttaPoints((pointX/diffX+1)*0.5, (pointY+1)*0.5);
		}
	}
}

/**
* Calculate following points of line using Euler.
*
* @param float x
* @param float y
*
*/
void Streamlines::getEulerPoints(float x, float y)
{
	float poX, poY, pkX, pkY; int j; float dx, dy, vel;
	float* points = new float[max*2];


	//Get corresponding x and y values from dataset
	//Calculate direction
	poX = x;
	poY = y;
	
	//Follow direction for max steps
	j = 0;
	for (int i = 0; i < max; i++)
	{
		dx = data->getChannel(0)->getValueNormPos(poX, poY);
		dy = data->getChannel(1)->getValueNormPos(poX, poY);
		vel = data->getChannel(7)->getValueNormPos(poX, poY);

		dx = dx/vel; dy = dy/vel;

		pkX = poX+dx*step;
		pkY = poY+dy*step;
		points[j] = poX; j++;
		points[j] = poY; j++;

		if (pkX < 0 || pkY < 0 || pkX > 1 || pkY > 1) 
		{
			points[j] = -1; j++;
			points[j] = -1;
			break;
		}

		poX = pkX; poY = pkY;
	}
	lines.push_back(points);
	points = new float[max*2];

	poX = x;
	poY = y;

	//Follow in opposite direction for max steps
	j = 0;
	for (int i = 0; i < max; i++)
	{
		dx = data->getChannel(0)->getValueNormPos(poX, poY);
		dy = data->getChannel(1)->getValueNormPos(poX, poY);
		vel = data->getChannel(7)->getValueNormPos(poX, poY);
		
		dx = dx/vel; dy = dy/vel;

		pkX = poX-dx*step;
		pkY = poY-dy*step;
		points[j] = poX; j++;
		points[j] = poY; j++;

		if (pkX < 0 || pkY < 0 || pkX > 1 || pkY > 1) 
		{
			points[j] = -1; j++;
			points[j] = -1;
			break;
		}

		poX = pkX; poY = pkY;
	}
	lines.push_back(points);
}

/**
* Calculate following points of line using Runge Kutta.
*
* @param float x
* @param float y
*
*/
void Streamlines::getRungeKuttaPoints(float x, float y)
{
	float poX, poY, pkX, pkY; int j; float dx, dy, vel;
	float* points = new float[max*2];


	//Get corresponding x and y values from dataset
	//Calculate direction
	poX = x;
	poY = y;
	
	//Follow direction for max steps
	j = 0;
	for (int i = 0; i < max; i++)
	{
		dx = data->getChannel(0)->getValueNormPos(poX, poY);
		dy = data->getChannel(1)->getValueNormPos(poX, poY);
		vel = data->getChannel(7)->getValueNormPos(poX, poY);
		
		dx = dx/vel; dy = dy/vel;

		pkX = poX+dx*step*0.5f;
		pkY = poY+dy*step*0.5f;

		points[j] = poX; j++;
		points[j] = poY; j++;

		if (pkX < 0 || pkY < 0 || pkX > 1 || pkY > 1) 
		{
			points[j] = -1; j++;
			points[j] = -1;
			break;
		}

		dx = data->getChannel(0)->getValueNormPos(pkX, pkY);
		dy = data->getChannel(1)->getValueNormPos(pkX, pkY);
		vel = data->getChannel(7)->getValueNormPos(pkX, pkY);

		dx = dx/vel; dy = dy/vel;

		pkX = poX+dx*step;
		pkY = poY+dy*step;

		if (pkX < 0 || pkY < 0 || pkX > 1 || pkY > 1) 
		{
			points[j] = -1; j++;
			points[j] = -1;
			break;
		}

		poX = pkX; poY = pkY;
	}
	lines.push_back(points);

	points = new float[max*2];

	poX = x;
	poY = y;

	//Follow in opposite direction for max steps
	j = 0;
	for (int i = 0; i < max; i++)
	{
		dx = data->getChannel(0)->getValueNormPos(poX, poY);
		dy = data->getChannel(1)->getValueNormPos(poX, poY);
		vel = data->getChannel(7)->getValueNormPos(poX, poY);
		
		dx = dx/vel; dy = dy/vel;

		pkX = poX-dx*step*0.5f;
		pkY = poY-dy*step*0.5f;

		points[j] = poX; j++;
		points[j] = poY; j++;

		if (pkX < 0 || pkY < 0 || pkX > 1 || pkY > 1) 
		{
			points[j] = -1; j++;
			points[j] = -1;
			break;
		}

		dx = data->getChannel(0)->getValueNormPos(pkX, pkY);
		dy = data->getChannel(1)->getValueNormPos(pkX, pkY);
		vel = data->getChannel(7)->getValueNormPos(pkX, pkY);

		dx = dx/vel; dy = dy/vel;

		pkX = poX-dx*step;
		pkY = poY-dy*step;

		if (pkX < 0 || pkY < 0 || pkX > 1 || pkY > 1) 
		{
			points[j] = -1; j++;
			points[j] = -1;
			break;
		}

		poX = pkX; poY = pkY;
	}
	lines.push_back(points);
}

/**
* Get stream lines.
* This function returns a vector which contains all points
* of the stream lines. 'formula' defines if Euler or Runge Kutta
* is used for calculations.
*
* @param formula
* @return std::vector<float*> streamlines
*
*/
std::vector<float*> Streamlines::getStreamLines(int formula)
{
	createGrid(formula);
	return lines;
}